# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).

from . import hooks, middleware, timeline_cursor


def setup_for_requests(timeline_factory=None, prefix=None):
    """Setup the hooks for timeline_django for request-based tracing.

    This will install all the Django hooks needed to populate the
    `Timeline` during requests. If you are running in an environment
    that is not request-based (manage.py commands, celery tasks) then
    you will need to set up the cursors and hooks in a different fashion.

    If `timeline_factory` is `None` then `timeline_django.middleware.get_timeline`
    will be used. This assumes that `timeline_django.middleware.TimelineMiddleware`
    is in the middleware for the Django project, and that the `timeline` wsgi setup
    is done so the middleware can work.

    If you aren't using the middleware, or aren't in a wsgi context, then you can
    provide your own timeline factory.
    """
    if timeline_factory is None:
        timeline_factory = middleware.get_timeline
    timeline_cursor.setup_timeline_cursor_hooks(timeline_factory, prefix=prefix)
    receiver = hooks.TimelineReceiver(timeline_factory)
    receiver.connect_to_signals()



# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).

import threading

from testtools import TestCase
from timeline import Timeline

from .. import middleware


class Request(object):

    def __init__(self, environ):
        self.environ = environ


class TimelineMiddlewareTests(TestCase):

    def setUp(self):
        super(TimelineMiddlewareTests, self).setUp()
        self.patch(middleware, '_thread_local', threading.local())

    def test_get_timeline_is_None(self):
        self.assertEqual(None, middleware.get_timeline())

    def test_middleware_doesnt_set_timeline_if_not_in_environ(self):
        mw = middleware.TimelineMiddleware()
        request = Request({})
        mw.process_request(request)
        self.assertEqual(None, middleware.get_timeline())

    def test_middleware_sets_timeline(self):
        mw = middleware.TimelineMiddleware()
        timeline = Timeline()
        request = Request({'timeline.timeline': timeline})
        mw.process_request(request)
        self.assertEqual(timeline, middleware.get_timeline())



# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).

import os
import unittest

def test_suite():
    os.environ['DJANGO_SETTINGS_MODULE'] = 'timeline_django.test_settings'
    module_names = [
        'timeline_django.tests.test_filters',
        'timeline_django.tests.test_hooks',
        'timeline_django.tests.test_middleware',
        'timeline_django.tests.test_timeline_cursor',
        ]
    loader = unittest.TestLoader()
    suite = loader.loadTestsFromNames(module_names)
    return suite



# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).

from datetime import timedelta
import threading

import django.core.signals
import django.db.backends.signals
import django.db.models.signals


class TimelineReceiver(object):
    """Converts django signals in to `Timeline` entries.

    To use this instantiate on instance, passing a callable that
    will return the timeline, and then call `connect_to_signals`.
    """

    INIT_CATEGORY = "model-init"
    SAVE_CATEGORY = "model-save"
    DELETE_CATEGORY = "model-delete"
    REQUEST_CATEGORY = "request"
    REQUEST_EXCEPTION_CATEGORY = "request-exception"
    CONNECTION_CREATED_CATEGORY = "connection-created"

    def __init__(self, timeline_factory):
        """Create a TimelineReceiver.

        :param timeline_factory: a callable that takes no arguments,
            and returns a `Timeline` object or `None`. This will
            be called each time a `Timeline` is needed in reponse
            to a signal.
        """
        self.timeline_factory = timeline_factory
        self._actions = threading.local()

    def _model_name(self, model_cls):
        return "%s.%s" % (model_cls.__module__, model_cls.__name__)

    def _action_name(self, category, sender):
        return "%s-%s" % (category, self._model_name(sender))

    def _handle_pre(self, category, sender, **kwargs):
        timeline = self.timeline_factory()
        if timeline is None:
            return
        action = timeline.start(category, self._model_name(sender),
                allow_nested=True)
        attr = self._action_name(category, sender)
        actions = getattr(self._actions, attr, [])
        actions.append(action)
        setattr(self._actions, attr, actions)

    def _handle_post(self, category, sender, **kwargs):
        action_name = self._action_name(category, sender)
        actions = getattr(self._actions, action_name, [])
        if not actions:
            raise AssertionError("post action called without pre action.")
        action = actions.pop()
        setattr(self._actions, action_name, actions)
        action.finish()

    def pre_init(self, sender, **kwargs):
        self._handle_pre(self.INIT_CATEGORY, sender, **kwargs)

    def post_init(self, sender, **kwargs):
        self._handle_post(self.INIT_CATEGORY, sender, **kwargs)

    def pre_save(self, sender, **kwargs):
        self._handle_pre(self.SAVE_CATEGORY, sender, **kwargs)

    def post_save(self, sender, **kwargs):
        self._handle_post(self.SAVE_CATEGORY, sender, **kwargs)

    def pre_delete(self, sender, **kwargs):
        self._handle_pre(self.DELETE_CATEGORY, sender, **kwargs)

    def post_delete(self, sender, **kwargs):
        self._handle_post(self.DELETE_CATEGORY, sender, **kwargs)

    # TODO: m2m_changed
    # TODO: 'using' kwarg in to the detail

    def request_started(self, sender, **kwargs):
        self._handle_pre(self.REQUEST_CATEGORY, sender, **kwargs)

    def request_finished(self, sender, **kwargs):
        self._handle_post(self.REQUEST_CATEGORY, sender, **kwargs)

    def _do_instantaneous_action(self, category, detail):
        timeline = self.timeline_factory()
        if timeline is None:
            return
        action = timeline.start(category, detail)
        action.duration = timedelta()

    def got_request_exception(self, sender, **kwargs):
        self._do_instantaneous_action(self.REQUEST_EXCEPTION_CATEGORY,
                self._model_name(sender))

    def connection_created(self, sender, **kwargs):
        connection = kwargs.get('connection', None)
        if connection is not None:
            connection_name = connection.alias
        else:
            connection_name = "(unknown)"
        self._do_instantaneous_action(self.CONNECTION_CREATED_CATEGORY,
                connection_name)

    def connect_to_signals(self):
        """Connect the callbacks to their corresponding signals."""
        django.db.models.signals.pre_init.connect(self.pre_init, weak=False)
        django.db.models.signals.post_init.connect(self.post_init, weak=False)
        django.db.models.signals.pre_save.connect(self.pre_save, weak=False)
        django.db.models.signals.post_save.connect(self.post_save, weak=False)
        django.db.models.signals.pre_delete.connect(self.pre_delete, weak=False)
        django.db.models.signals.post_delete.connect(self.post_delete, weak=False)
        django.core.signals.request_started.connect(self.request_started, weak=False)
        django.core.signals.request_finished.connect(self.request_finished, weak=False)
        django.core.signals.got_request_exception.connect(self.got_request_exception, weak=False)
        django.db.backends.signals.connection_created.connect(self.connection_created, weak=False)



# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).

import oops
import oops_timeline
from testtools import TestCase
import timeline

from .. import filters


class FilterSessionInfoTests(TestCase):

    def test_leaves_most_queries_be(self):
        query = 'some query'
        result = filters.filter_session_info(query)
        self.assertEqual(query, result)

    def test_redacts_django_session(self):
        query = 'select * from django_session'
        result = filters.filter_session_info(query)
        self.assertEqual('select <session query redacted>', result)


class FilterUserInfoTests(TestCase):

    def test_leaves_most_queries_be(self):
        query = 'some query'
        result = filters.filter_user_info(query)
        self.assertEqual(query, result)

    def test_redacts_django_session(self):
        query = 'select * from auth_user'
        result = filters.filter_user_info(query)
        self.assertEqual('select <user query redacted>', result)


class FilterSessionQueryTests(TestCase):

    def test_does_nothing_no_timeline(self):
        report = {}
        context = {}
        filters.filter_session_query(report, context)
        self.assertEqual({}, report)
        self.assertEqual({}, context)

    def test_filter_session_query(self):
        query = 'select * from django_session'
        test_timeline = [(0, 0, 'some-category', query, 'other')]
        report = dict(timeline=test_timeline)
        filters.filter_session_query(report, {})
        new_timeline = report['timeline']
        self.assertEqual(
                [(0, 0, 'some-category', filters.filter_session_info(query),
                    'other')],
                new_timeline)


class FilterUserQueryTests(TestCase):

    def test_does_nothing_no_timeline(self):
        report = {}
        context = {}
        filters.filter_user_query(report, context)
        self.assertEqual({}, report)
        self.assertEqual({}, context)

    def test_filter_user_query(self):
        query = 'select * from auth_user'
        test_timeline = [(0, 0, 'some-category', query, 'other')]
        report = dict(timeline=test_timeline)
        filters.filter_user_query(report, {})
        new_timeline = report['timeline']
        self.assertEqual(
                [(0, 0, 'some-category', filters.filter_user_info(query),
                    'other')],
                new_timeline)


class FakeConfig(object):

    def __init__(self):
        self.on_create = []


class InstallHooksTests(TestCase):

    def test_installs_filter_session_query(self):
        config = FakeConfig()
        filters.install_hooks(config)
        self.assertIn(filters.filter_session_query, config.on_create)

    def test_installs_filter_user_query(self):
        config = FakeConfig()
        filters.install_hooks(config)
        self.assertIn(filters.filter_user_query, config.on_create)


class IntegrationTests(TestCase):

    def test_publishing_oops_redacts_timeline(self):
        config = oops.Config()
        oops_timeline.install_hooks(config)
        filters.install_hooks(config)
        user_query = 'select * from auth_user'
        session_query = 'select * from django_session'
        test_timeline = timeline.Timeline()
        test_timeline.start('some-category', user_query).finish()
        test_timeline.start('some-other-category', session_query).finish()
        test_context = dict(timeline=test_timeline)
        test_oops = config.create(test_context)
        self.assertEqual(
            [filters.filter_user_info(user_query),
             filters.filter_session_info(session_query),
            ],
            [action[3] for action in test_oops['timeline']])



# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).

import threading


TIMELINE_VAR = 'timeline'


_thread_local = threading.local()


def get_timeline():
    """Get the timeline associated with the current request, if any.

    If `TimelineMiddleware` saved a `Timeline` object at the start of
    the request calling this function will return it. If there isn't
    a saved timeline then `None` will be returned.
    """
    return getattr(_thread_local, TIMELINE_VAR, None)


class TimelineMiddleware(object):
    """Save a `Timeline` as a thread local.

    This middleware looks in `request.environ` for a `timeline.timeline`
    key, and if it is present it stores it in a thread local variable,
    so that `get_timeline` will return it when later called in that
    thread.
    """

    TIMELINE_ENVIRON_KEY = 'timeline.timeline'

    def process_request(self, request):
        if self.TIMELINE_ENVIRON_KEY in request.environ:
            setattr(_thread_local, TIMELINE_VAR,
                    request.environ[self.TIMELINE_ENVIRON_KEY])



# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).

import django.db
import django.core.signals
from django.dispatch import Signal
from testtools import TestCase
from testtools.matchers import StartsWith, EndsWith
from timeline import Timeline

from ..timeline_cursor import (
    TimelineCursor,
    TimelineCursorInserter,
    )


class TestCursor(object):

    def __init__(self):
        self.calls = []
        self.many_calls = []

    def execute(self, sql, params):
        self.calls.append((sql, params))

    def executemany(self, sql, param_list):
        self.many_calls.append((sql, param_list))


class ExpectedException(Exception):
    pass


class RaisingCursor(TestCursor):

    def execute(self, sql, params=None):
        super(RaisingCursor, self).execute(sql, params=params)
        raise ExpectedException()


class ManyRaisingCursor(TestCursor):

    def executemany(self, sql, param_list):
        super(ManyRaisingCursor, self).executemany(sql, param_list)
        raise ExpectedException()


class TestDbOps(object):

    def __init__(self, db):
        self.db = db

    def last_executed_query(self, cursor, sql, params):
        if cursor != self.db.cursor:
            raise AssertionError("Called with wrong cursor: %s" % cursor)
        if len(cursor.calls) < 1:
            raise AssertionError("Called before execute")
        if sql != cursor.calls[-1][0]:
            raise AssertionError("Called with wrong sql: %s" % sql)
        if params != cursor.calls[-1][1]:
            raise AssertionError("Called with wrong params: %s" % params)
        return self.db.last_query


class TestDatabase(object):

    def __init__(self, alias, cursor, last_query):
        self.alias = alias
        self.cursor = cursor
        self.last_query = last_query
        self.ops = TestDbOps(self)


class TimelineCursorTests(TestCase):

    _not_passed = object()

    def get_cursor(self, inner_cursor=None, db=None, alias=None,
            timeline_factory=None, prefix=_not_passed):
        if inner_cursor is None:
            inner_cursor = TestCursor()
        if alias is None:
            alias = self.getUniqueString()
        if db is None:
            db = TestDatabase(alias, inner_cursor, self.getUniqueString())
        if timeline_factory is None:
            timeline = Timeline()
            timeline_factory = lambda: timeline
        if prefix is self._not_passed:
            prefix = self.getUniqueString()
        return TimelineCursor(inner_cursor, db, timeline_factory,
                prefix=prefix)

    def test_execute_with_no_timeline(self):
        inner_cursor = TestCursor()
        cursor = self.get_cursor(inner_cursor=inner_cursor,
                timeline_factory=lambda: None)
        sql = 'SELECT * from foo;'
        params = (self.getUniqueString(),)
        # Shouldn't crash trying to use the None returned from the
        # timeline_factory
        cursor.execute(sql, params)
        self.assertEqual([(sql, params)], inner_cursor.calls)

    def test_execute_creates_action(self):
        timeline = Timeline()
        cursor = self.get_cursor(timeline_factory=lambda: timeline)
        cursor.execute(self.getUniqueString())
        self.assertEqual(1, len(timeline.actions))

    def test_execute_finishes_action(self):
        timeline = Timeline()
        cursor = self.get_cursor(timeline_factory=lambda: timeline)
        cursor.execute(self.getUniqueString())
        self.assertEqual(1, len(timeline.actions))
        # Duration == None indicates that the action is unfinished
        self.assertNotEqual(None, timeline.actions[0].duration)

    def test_execute_calls_cursor(self):
        inner_cursor = TestCursor()
        cursor = self.get_cursor(inner_cursor=inner_cursor)
        sql = 'SELECT * from foo;'
        params=(1, 2,)
        cursor.execute(sql, params=params)
        self.assertEqual(1, len(inner_cursor.calls))
        self.assertEqual((sql, params), inner_cursor.calls[0])

    def test_execute_finishes_action_even_if_inner_cursor_raises(self):
        inner_cursor = RaisingCursor()
        timeline = Timeline()
        cursor = self.get_cursor(inner_cursor=inner_cursor,
                timeline_factory=lambda: timeline)
        self.assertRaises(ExpectedException,
                cursor.execute, self.getUniqueString())
        self.assertEqual(1, len(timeline.actions))
        # Duration == None indicates that the action is unfinished
        self.assertNotEqual(None, timeline.actions[0].duration)

    def test_execute_updates_detail_based_on_last_executed_query(self):
        timeline = Timeline()
        alias = self.getUniqueString()
        inner_cursor = TestCursor()
        last_query = self.getUniqueString()
        db = TestDatabase(alias, inner_cursor, last_query)
        cursor = self.get_cursor(inner_cursor=inner_cursor, db=db,
                timeline_factory=lambda: timeline)
        cursor.execute(self.getUniqueString())
        self.assertEqual(1, len(timeline.actions))
        self.assertEqual(last_query, timeline.actions[0].detail)

    def test_execute_action_starts_with_prefix(self):
        prefix = self.getUniqueString()
        timeline = Timeline()
        cursor = self.get_cursor(timeline_factory=lambda: timeline,
                prefix=prefix)
        cursor.execute(self.getUniqueString())
        self.assertThat(timeline.actions[0].category, StartsWith(prefix))

    def test_execute_default_prefix(self):
        timeline = Timeline()
        cursor = self.get_cursor(timeline_factory=lambda: timeline, prefix=None)
        cursor.execute(self.getUniqueString())
        self.assertThat(timeline.actions[0].category, StartsWith('SQL-'))

    def test_execute_action_ends_with_connection_name(self):
        connection_name = self.getUniqueString()
        timeline = Timeline()
        cursor = self.get_cursor(alias=connection_name,
                timeline_factory=lambda: timeline)
        cursor.execute(self.getUniqueString())
        self.assertThat(timeline.actions[0].category, EndsWith(connection_name))

    def test_executemany_with_no_timeline(self):
        inner_cursor = TestCursor()
        cursor = self.get_cursor(inner_cursor=inner_cursor,
                timeline_factory=lambda: None)
        sql = 'SELECT * from foo;'
        param_list = [(self.getUniqueString(),)]
        # Shouldn't crash trying to use the None returned from the
        # timeline_factory
        cursor.executemany(sql, param_list)
        self.assertEqual([(sql, param_list)], inner_cursor.many_calls)

    def test_executemany_creates_action(self):
        timeline = Timeline()
        cursor = self.get_cursor(timeline_factory=lambda: timeline)
        cursor.executemany(self.getUniqueString())
        self.assertEqual(1, len(timeline.actions))

    def test_executemany_finishes_action(self):
        timeline = Timeline()
        cursor = self.get_cursor(timeline_factory=lambda: timeline)
        cursor.executemany(self.getUniqueString())
        self.assertEqual(1, len(timeline.actions))
        # Duration == None indicates that the action is unfinished
        self.assertNotEqual(None, timeline.actions[0].duration)

    def test_executemany_calls_cursor(self):
        inner_cursor = TestCursor()
        cursor = self.get_cursor(inner_cursor=inner_cursor)
        sql = 'SELECT * from foo;'
        param_list=[(1, 2,)]
        cursor.executemany(sql, param_list)
        self.assertEqual([(sql, param_list)], inner_cursor.many_calls)

    def test_executemany_finishes_action_even_if_inner_cursor_raises(self):
        inner_cursor = ManyRaisingCursor()
        timeline = Timeline()
        cursor = self.get_cursor(inner_cursor=inner_cursor,
                timeline_factory=lambda: timeline)
        self.assertRaises(ExpectedException,
                cursor.executemany, self.getUniqueString())
        self.assertEqual(1, len(timeline.actions))
        # Duration == None indicates that the action is unfinished
        self.assertNotEqual(None, timeline.actions[0].duration)

    def test_executemany_updates_detail_based_on_sql_and_param_list(self):
        timeline = Timeline()
        cursor = self.get_cursor(timeline_factory=lambda: timeline)
        sql = self.getUniqueString()
        param_list = [(self.getUniqueString(),), (self.getUniqueString(),)]
        cursor.executemany(sql, param_list)
        self.assertEqual(1, len(timeline.actions))
        self.assertEqual(sql + " " + str(param_list) , timeline.actions[0].detail)

    def test_executemany_action_starts_with_prefix(self):
        prefix = self.getUniqueString()
        timeline = Timeline()
        cursor = self.get_cursor(timeline_factory=lambda: timeline,
                prefix=prefix)
        cursor.executemany(self.getUniqueString())
        self.assertThat(timeline.actions[0].category, StartsWith(prefix))

    def test_executemany_default_prefix(self):
        timeline = Timeline()
        cursor = self.get_cursor(timeline_factory=lambda: timeline, prefix=None)
        cursor.executemany(self.getUniqueString())
        self.assertThat(timeline.actions[0].category, StartsWith('SQL-'))

    def test_executemany_action_ends_with_connection_name(self):
        connection_name = self.getUniqueString()
        timeline = Timeline()
        cursor = self.get_cursor(alias=connection_name,
                timeline_factory=lambda: timeline)
        cursor.executemany(self.getUniqueString())
        self.assertThat(timeline.actions[0].category, EndsWith(connection_name))


class TestConnection(object):

    def __init__(self, cursor=None):
        self.alias = "an-alias"
        self.cursor = lambda: cursor


class InsertTimelineCursorsTests(TestCase):

    def make_and_patch_connection(self, cursor=None, connection_name=None):
        """Patch the django.db.connections dict with a test double.

        In order to isolate the tests from the environment we
        monkey patch django.db.connections.

        This dict is the mapping from connection name to
        connection (`DatabaseWrapper`.)

        :param cursor: the cursor to use in the mock Connection.
        :param connection_name: the name the connection should have,
            or None for the method to generate an arbitrary unique
            name.
        """
        connection = TestConnection(cursor=cursor)
        if connection_name is None:
            connection_name = self.getUniqueString()
        connections = {connection_name: connection}
        self.patch(django.db, 'connections', connections)
        return connection

    def get_cursor(self):
        return django.db.connections.values()[0].cursor()

    def test_wraps_one_connection(self):
        self.make_and_patch_connection()
        inserter = TimelineCursorInserter(lambda: None)
        inserter.insert_timeline_cursors()
        self.assertEqual(1, len(django.db.connections))
        self.assertIsInstance(self.get_cursor(), TimelineCursor)

    def test_wraps_two_conections(self):
        connections = [TestConnection(), TestConnection()]
        connections = dict(foo=connections[0], bar=connections[1])
        self.patch(django.db, 'connections', connections)
        inserter = TimelineCursorInserter(lambda: None)
        inserter.insert_timeline_cursors()
        self.assertEqual(2, len(connections))
        self.assertIsInstance(connections.values()[0].cursor(), TimelineCursor)
        self.assertIsInstance(connections.values()[1].cursor(), TimelineCursor)

    def test_wraps_original_cursor(self):
        cursor = TestCursor()
        self.make_and_patch_connection(cursor=cursor)
        inserter = TimelineCursorInserter(lambda: None)
        inserter.insert_timeline_cursors()
        self.assertEqual(1, len(django.db.connections))
        self.assertEqual(cursor, self.get_cursor().cursor)

    def test_sets_db(self):
        connection = self.make_and_patch_connection()
        inserter = TimelineCursorInserter(lambda: None)
        inserter.insert_timeline_cursors()
        self.assertEqual(1, len(django.db.connections))
        self.assertEqual(connection, self.get_cursor().db)

    def test_sets_timeline_factory(self):
        self.make_and_patch_connection()
        timeline_factory = lambda: None
        inserter = TimelineCursorInserter(timeline_factory)
        inserter.insert_timeline_cursors()
        self.assertEqual(1, len(django.db.connections))
        self.assertEqual(timeline_factory, self.get_cursor().timeline_factory)

    def test_sets_prefix(self):
        self.make_and_patch_connection()
        prefix = self.getUniqueString()
        inserter = TimelineCursorInserter(lambda: None, prefix=prefix)
        inserter.insert_timeline_cursors()
        self.assertEqual(1, len(django.db.connections))
        self.assertEqual(prefix, self.get_cursor().prefix)

    def test_wraps_only_once(self):
        # Calling twice from same thread doesn't lead to double wrapping
        cursor = TestCursor()
        self.make_and_patch_connection(cursor=cursor)
        inserter = TimelineCursorInserter(lambda: None)
        inserter.insert_timeline_cursors()
        inserter.insert_timeline_cursors()
        self.assertEqual(1, len(django.db.connections))
        # Check that the original connection is now the cursor
        # attribute of the wrapped cursor. If calling
        # insert_timeline_cursors twice wrapped twice then the
        # attribute would be another cursor wrapper that wraps
        # the original cursor.
        self.assertEqual(cursor, self.get_cursor().cursor)

    def patch_signal(self, module, signal_name):
        signal = Signal()
        self.patch(module, signal_name, signal)
        return signal

    def test_register_signals_registers_request_started(self):
        signal = self.patch_signal(django.core.signals, 'request_started')
        inserter = TimelineCursorInserter(lambda: None)
        inserter.connect_to_request_signals()
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(inserter._request_cb, signal.receivers[0][1])

    def test_request_cb_wraps_one_connection(self):
        self.make_and_patch_connection()
        inserter = TimelineCursorInserter(lambda: None)
        inserter._request_cb(self, something=self.getUniqueString())
        self.assertEqual(1, len(django.db.connections))
        self.assertIsInstance(self.get_cursor(), TimelineCursor)



# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).

import os

from setuptools import setup, find_packages

setup(
    name='timeline_django',
    version='0.0.1',
    packages=find_packages(),
    include_package_data=True,
    maintainer='Launchpad developers',
    maintainer_email='launchpad-dev@lists.launchpad.net',
    description='Insert Django database queries in to a Timeline',
    long_description=open(os.path.join(os.path.dirname(__file__), 'README')).read(),
    license='LGPLv3',
    url='http://launchpad.net/python-timeline-django',
    download_url='https://launchpad.net/python-timeline-django/+download',
    test_suite='timeline_django.tests',
    install_requires = [
        'Django',
        ],
    tests_require = [
        'testtools',
        'timeline',
        'oops',
        'oops-timeline',
        ],
    # Auto-conversion to Python 3.
    use_2to3=True,
    )



# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).


SESSION_TABLE_NAME = 'django_session'
USER_TABLE_NAME = 'auth_user'


def _filter_table(query, table, reason):
    if table in query:
        return query.split(" ", 1)[0] + " <%s query redacted>" % reason
    return query


def filter_session_info(query):
    return _filter_table(query, SESSION_TABLE_NAME, 'session')


def filter_user_info(query):
    return _filter_table(query, USER_TABLE_NAME, 'user')


def _filter_query(report, filter_fn):
    timeline = report.get('timeline')
    if timeline is None:
        return
    new_timeline = []
    for event in timeline:
        start, end, category, detail = event[:4]
        detail = filter_fn(detail)
        new_timeline.append((start, end, category, detail) + event[4:])
    report['timeline'] = new_timeline


def filter_session_query(report, context):
    """Filter out queries about the session from the timeline."""
    _filter_query(report, filter_session_info)


def filter_user_query(report, context):
    """Filter out queries about the user from the timeline."""
    _filter_query(report, filter_user_info)


def install_hooks(config):
    """Install on_create hooks in the config to sanitise queries.

    This should be called after the `oops_timeline` hooks are instaled.

    :param config: the `oops.Config` to install the hooks in to.
    """
    config.on_create.append(filter_session_query)
    config.on_create.append(filter_user_query)



# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).

from datetime import timedelta

import django.core.signals
import django.db.backends.signals
import django.db.models.signals
from django.dispatch import Signal
from testtools import TestCase
from timeline import Timeline

from ..hooks import TimelineReceiver


class TestModel(object):

    # Must be the FQPN of the class to correspond with
    # TimelineReceiver._model_name
    NAME = '%s.%s' % (__name__, 'TestModel')


class TestConnection(object):

    def __init__(self, alias=None):
        if alias is None:
            alias = "foo"
        self.alias = alias


class TimelineReceiverTests(TestCase):

    def get_timeline_receiver(self):
        timeline = Timeline()
        return TimelineReceiver(lambda: timeline), timeline

    def test_pre_init_with_no_timeline(self):
        receiver = TimelineReceiver(lambda: None)
        # Mustn't crash due to trying to use a None timeline
        receiver.pre_init(TestModel)

    def test_pre_init_adds_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_init(TestModel)
        self.assertEqual(1, len(timeline.actions))

    def test_pre_init_sets_model_init_as_category(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_init(TestModel)
        # -start is appended as this is a nestable action
        self.assertEqual('model-init-start', timeline.actions[0].category)

    def test_pre_init_sets_model_name_as_detail(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_init(TestModel)
        self.assertEqual(TestModel.NAME, timeline.actions[0].detail)

    def test_pre_init_finishes_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_init(TestModel)
        # If this passes it means the action was completed as an
        # instantaneous action. This is so nesting can be done,
        # as the action may well trigger SQL
        self.assertEqual(timedelta(), timeline.actions[0].duration)

    def test_post_init_with_no_pre_init(self):
        receiver, timeline = self.get_timeline_receiver()
        self.assertRaises(AssertionError, receiver.post_init, TestModel)

    def test_post_init_adds_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_init(TestModel)
        receiver.post_init(TestModel)
        self.assertEqual(2, len(timeline.actions))

    def test_post_init_sets_model_init_as_category(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_init(TestModel)
        receiver.post_init(TestModel)
        # -finish is appended as this is a nestable action
        self.assertEqual('model-init-stop', timeline.actions[-1].category)

    def test_post_init_sets_model_name_as_detail(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_init(TestModel)
        receiver.post_init(TestModel)
        self.assertEqual(TestModel.NAME, timeline.actions[-1].detail)

    def test_post_init_finishes_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_init(TestModel)
        receiver.post_init(TestModel)
        self.assertEqual(timedelta(), timeline.actions[-1].duration)

    def test_nested_init(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_init(TestModel)
        receiver.pre_init(TestModel)
        receiver.post_init(TestModel)
        receiver.post_init(TestModel)
        self.assertEqual(4, len(timeline.actions))


    def test_pre_save_with_no_timeline(self):
        receiver = TimelineReceiver(lambda: None)
        # Mustn't crash due to trying to use a None timeline
        receiver.pre_save(TestModel)

    def test_pre_save_adds_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_save(TestModel)
        self.assertEqual(1, len(timeline.actions))

    def test_pre_save_sets_model_save_as_category(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_save(TestModel)
        # -start is appended as this is a nestable action
        self.assertEqual('model-save-start', timeline.actions[0].category)

    def test_pre_save_sets_model_name_as_detail(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_save(TestModel)
        self.assertEqual(TestModel.NAME, timeline.actions[0].detail)

    def test_pre_save_finishes_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_save(TestModel)
        # If this passes it means the action was completed as an
        # instantaneous action. This is so nesting can be done,
        # as the action may well trigger SQL
        self.assertEqual(timedelta(), timeline.actions[0].duration)

    def test_post_save_with_no_pre_save(self):
        receiver, timeline = self.get_timeline_receiver()
        self.assertRaises(AssertionError, receiver.post_save, TestModel)

    def test_post_save_adds_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_save(TestModel)
        receiver.post_save(TestModel)
        self.assertEqual(2, len(timeline.actions))

    def test_post_save_sets_model_save_as_category(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_save(TestModel)
        receiver.post_save(TestModel)
        # -finish is appended as this is a nestable action
        self.assertEqual('model-save-stop', timeline.actions[-1].category)

    def test_post_save_sets_model_name_as_detail(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_save(TestModel)
        receiver.post_save(TestModel)
        self.assertEqual(TestModel.NAME, timeline.actions[-1].detail)

    def test_post_save_finishes_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_save(TestModel)
        receiver.post_save(TestModel)
        # The save action is nestable, meaning it is represented as
        # two instantaneous actions, so check that the second
        # action has no duration.
        self.assertEqual(timedelta(), timeline.actions[-1].duration)

    def test_nested_save(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_save(TestModel)
        receiver.pre_save(TestModel)
        receiver.post_save(TestModel)
        receiver.post_save(TestModel)
        self.assertEqual(4, len(timeline.actions))


    def test_pre_delete_with_no_timeline(self):
        receiver = TimelineReceiver(lambda: None)
        # Mustn't crash due to trying to use a None timeline
        receiver.pre_delete(TestModel)

    def test_pre_delete_adds_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_delete(TestModel)
        self.assertEqual(1, len(timeline.actions))

    def test_pre_delete_sets_model_delete_as_category(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_delete(TestModel)
        # -start is appended as this is a nestable action
        self.assertEqual('model-delete-start', timeline.actions[0].category)

    def test_pre_delete_sets_model_name_as_detail(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_delete(TestModel)
        self.assertEqual(TestModel.NAME, timeline.actions[0].detail)

    def test_pre_delete_finishes_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_delete(TestModel)
        # If this passes it means the action was completed as an
        # instantaneous action. This is so nesting can be done,
        # as the action may well trigger SQL
        self.assertEqual(timedelta(), timeline.actions[0].duration)

    def test_post_delete_with_no_pre_delete(self):
        receiver, timeline = self.get_timeline_receiver()
        self.assertRaises(AssertionError, receiver.post_delete, TestModel)

    def test_post_delete_adds_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_delete(TestModel)
        receiver.post_delete(TestModel)
        self.assertEqual(2, len(timeline.actions))

    def test_post_delete_sets_model_delete_as_category(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_delete(TestModel)
        receiver.post_delete(TestModel)
        # -finish is appended as this is a nestable action
        self.assertEqual('model-delete-stop', timeline.actions[-1].category)

    def test_post_delete_sets_model_name_as_detail(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_delete(TestModel)
        receiver.post_delete(TestModel)
        self.assertEqual(TestModel.NAME, timeline.actions[-1].detail)

    def test_post_delete_finishes_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_delete(TestModel)
        receiver.post_delete(TestModel)
        self.assertEqual(timedelta(), timeline.actions[-1].duration)

    def test_nested_delete(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.pre_delete(TestModel)
        receiver.pre_delete(TestModel)
        receiver.post_delete(TestModel)
        receiver.post_delete(TestModel)
        self.assertEqual(4, len(timeline.actions))


    def test_request_started_with_no_timeline(self):
        receiver = TimelineReceiver(lambda: None)
        # Mustn't crash due to trying to use a None timeline
        receiver.request_started(TestModel)

    def test_request_started_adds_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.request_started(TestModel)
        self.assertEqual(1, len(timeline.actions))

    def test_request_started_sets_request_as_category(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.request_started(TestModel)
        # -start is appended as this is a nestable action
        self.assertEqual('request-start', timeline.actions[0].category)

    def test_request_started_sets_sender_as_detail(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.request_started(TestModel)
        self.assertEqual(TestModel.NAME, timeline.actions[0].detail)

    def test_request_started_finishes_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.request_started(TestModel)
        # If this passes it means the action was completed as an
        # instantaneous action. This is so nesting can be done,
        # as the action may well trigger SQL
        self.assertEqual(timedelta(), timeline.actions[0].duration)

    def test_request_finished_with_no_request_started(self):
        receiver, timeline = self.get_timeline_receiver()
        self.assertRaises(AssertionError, receiver.request_finished, TestModel)

    def test_request_finished_adds_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.request_started(TestModel)
        receiver.request_finished(TestModel)
        self.assertEqual(2, len(timeline.actions))

    def test_request_finished_sets_request_as_category(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.request_started(TestModel)
        receiver.request_finished(TestModel)
        # -finish is appended as this is a nestable action
        self.assertEqual('request-stop', timeline.actions[-1].category)

    def test_request_finished_sets_model_name_as_detail(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.request_started(TestModel)
        receiver.request_finished(TestModel)
        self.assertEqual(TestModel.NAME, timeline.actions[-1].detail)

    def test_request_finished_finishes_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.request_started(TestModel)
        receiver.request_finished(TestModel)
        self.assertEqual(timedelta(), timeline.actions[-1].duration)

    def test_nested_request(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.request_started(TestModel)
        receiver.request_started(TestModel)
        receiver.request_finished(TestModel)
        receiver.request_finished(TestModel)
        self.assertEqual(4, len(timeline.actions))


    def test_got_request_exception_with_no_timeline(self):
        receiver = TimelineReceiver(lambda: None)
        # Mustn't crash due to trying to use a None timeline
        receiver.got_request_exception(TestModel)

    def test_got_request_exception_adds_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.got_request_exception(TestModel)
        self.assertEqual(1, len(timeline.actions))

    def test_got_request_exception_sets_request_exception_as_category(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.got_request_exception(TestModel)
        self.assertEqual('request-exception', timeline.actions[0].category)

    def test_got_request_exception_sets_sender_as_detail(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.got_request_exception(TestModel)
        self.assertEqual(TestModel.NAME, timeline.actions[0].detail)

    def test_got_request_exception_finishes_action(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.got_request_exception(TestModel)
        self.assertEqual(timedelta(), timeline.actions[0].duration)


    def test_connection_created_with_no_timeline(self):
        receiver = TimelineReceiver(lambda: None)
        connection = TestConnection()
        # Mustn't crash due to trying to use a None timeline
        receiver.connection_created(TestModel, connection=connection)

    def test_connection_created_adds_action(self):
        receiver, timeline = self.get_timeline_receiver()
        connection = TestConnection()
        receiver.connection_created(TestModel, connection=connection)
        self.assertEqual(1, len(timeline.actions))

    def test_connection_created_sets_connection_created_as_category(self):
        receiver, timeline = self.get_timeline_receiver()
        connection = TestConnection()
        receiver.connection_created(TestModel, connection=connection)
        self.assertEqual('connection-created', timeline.actions[0].category)

    def test_connection_created_sets_connection_name_as_detail(self):
        receiver, timeline = self.get_timeline_receiver()
        connection_name = self.getUniqueString()
        connection = TestConnection(alias=connection_name)
        receiver.connection_created(TestModel, connection=connection)
        self.assertEqual(connection_name, timeline.actions[0].detail)

    def test_connection_created_sets_unknown_if_connection_missing(self):
        receiver, timeline = self.get_timeline_receiver()
        receiver.connection_created(TestModel)
        self.assertEqual("(unknown)", timeline.actions[0].detail)

    def test_connection_created_finishes_action(self):
        receiver, timeline = self.get_timeline_receiver()
        connection = TestConnection()
        receiver.connection_created(TestModel, connection=connection)
        self.assertEqual(timedelta(), timeline.actions[0].duration)


    def patch_signal(self, module, signal_name):
        signal = Signal()
        self.patch(module, signal_name, signal)
        return signal

    def register_signal(self, module, signal_name):
        signal = self.patch_signal(module, signal_name)
        receiver, timeline = self.get_timeline_receiver()
        receiver.connect_to_signals()
        return signal, receiver

    def test_register_signals_registers_pre_init(self):
        signal, receiver = self.register_signal(django.db.models.signals, 'pre_init')
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(receiver.pre_init, signal.receivers[0][1])

    def test_register_signals_registers_post_init(self):
        signal, receiver = self.register_signal(django.db.models.signals, 'post_init')
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(receiver.post_init, signal.receivers[0][1])

    def test_register_signals_registers_pre_save(self):
        signal, receiver = self.register_signal(django.db.models.signals, 'pre_save')
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(receiver.pre_save, signal.receivers[0][1])

    def test_register_signals_registers_post_save(self):
        signal, receiver = self.register_signal(django.db.models.signals, 'post_save')
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(receiver.post_save, signal.receivers[0][1])

    def test_register_signals_registers_pre_delete(self):
        signal, receiver = self.register_signal(django.db.models.signals, 'pre_delete')
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(receiver.pre_delete, signal.receivers[0][1])

    def test_register_signals_registers_post_delete(self):
        signal, receiver = self.register_signal(django.db.models.signals, 'post_delete')
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(receiver.post_delete, signal.receivers[0][1])

    def test_register_signals_registers_request_started(self):
        signal, receiver = self.register_signal(django.core.signals, 'request_started')
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(receiver.request_started, signal.receivers[0][1])

    def test_register_signals_registers_request_finished(self):
        signal, receiver = self.register_signal(django.core.signals, 'request_finished')
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(receiver.request_finished, signal.receivers[0][1])

    def test_register_signals_registers_got_request_exception(self):
        signal, receiver = self.register_signal(django.core.signals, 'got_request_exception')
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(receiver.got_request_exception, signal.receivers[0][1])

    def test_register_signals_registers_connection_created(self):
        signal, receiver = self.register_signal(django.db.backends.signals, 'connection_created')
        self.assertEqual(1, len(signal.receivers))
        self.assertEqual(receiver.connection_created, signal.receivers[0][1])



# Copyright (c) 2012, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 only.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Lesser General Public License version 3 (see the file LICENSE).

import threading

from django import db
from django.db.backends.util import CursorWrapper


class TimelineCursor(CursorWrapper):
    """A wrapper around a Django db cursor that records actions in a `Timeline`.

    Wrapping a Django db cursor with a `TimelineCursor` allows for all the
    executed queries to be recorded in a timeline.
    """

    def __init__(self, cursor, db, timeline_factory, prefix=None):
        """Create a `TimelineCursor`.

        :param cursor: the cursor being wrapped.
        :param db: the `DatabaseWrapper` that the cursor is from.
        :param timeline_factory: a callable that takes no arguments
            and returns a `Timeline` instance or None. This will be
            called whenever a new query is made to get the timeline
            to record the action on. If it returns None the action
            won't be recorded.
        :param prefix: the prefix to use for actions inserted in to
            the timeline. If None the default of 'SQL-' will be used.
        """
        super(TimelineCursor, self).__init__(cursor, db)
        self.cursor = cursor
        self.db = db
        self.connection_name = self.db.alias
        self.timeline_factory = timeline_factory
        self.prefix = prefix
        if self.prefix is None:
            self.prefix = 'SQL-'

    def execute(self, sql, params=()):
        action = None
        timeline = self.timeline_factory()
        if timeline is not None:
            action = timeline.start(self.prefix+self.connection_name,
                    sql + " " + str(params))
        try:
            self.cursor.execute(sql, params)
        finally:
            if timeline is not None:
                action.detail = self.db.ops.last_executed_query(self.cursor,
                        sql, params)
                action.finish()

    def executemany(self, sql, param_list=[]):
        action = None
        timeline = self.timeline_factory()
        if timeline is not None:
            action = timeline.start(self.prefix+self.connection_name,
                    sql + " " + str(param_list))
        try:
            self.cursor.executemany(sql, param_list)
        finally:
            if timeline is not None:
                action.finish()


class TimelineCursorInserter(object):
    """Inserts `TimelineCursor`s once per-thread.

    Calling `insert_timeline_cursors` will insert a `TimelineCursor` in to the
    Django db machinery, but skip if this instance has already done that before
    in this thread.

    You can trigger that method such that it will run at least once in each
    thread to ensure that all cursors are wrapped in a `TimelineCursor`.

    A common way to do this is to do it at the start of every request. To
    aid with doing that you can call `connect_to_request_signals` which
    well setup a receiver for `django.core.signals.request_started` to
    call `insert_timeline_cursors`.
    """

    def __init__(self, timeline_factory, prefix=None):
        """Make a `TimelineCursorInserter`.

        :param timeline_factory: a callable that takes no arguments
            and returns a `Timeline` instance or None.
        :param prefix: the prefix to use for the entries recorded
            in the `Timeline`, or None to use the default.
        """
        self._cursor_wrapped = threading.local()
        self.timeline_factory = timeline_factory
        self.prefix = prefix

    def insert_timeline_cursors(self):
        """Insert `TimelineCursor`s in to the db machinery.

        No action will be taken if the method has already been called
        in the current thread.
        """
        if not getattr(self._cursor_wrapped, 'done', False):
            for connection_name in db.connections:
                connection = db.connections[connection_name]
                connection.cursor = self._make_timeline_cursor_wrapper(
                        connection)
            setattr(self._cursor_wrapped, 'done', True)

    def _make_timeline_cursor_wrapper(self, connection):
        _orig_cursor_fn = connection.cursor
        def make_timeline_cursor():
            cursor = _orig_cursor_fn()
            return TimelineCursor(cursor, connection, self.timeline_factory,
                    prefix=self.prefix)
        return make_timeline_cursor

    def _request_cb(self, sender, **kwargs):
        self.insert_timeline_cursors()

    def connect_to_request_signals(self):
        """Connect to the `request_started` signal to insert the cursors.

        After calling this every time the `request_started` signal is
        triggered `insert_timeline_cursors` will be called.
        """
        from django.core.signals import request_started
        request_started.connect(self._request_cb, weak=False)


def setup_timeline_cursor_hooks(timeline_factory, prefix=None):
    """Setup the hooks necessary to insert `TimelineCursor`s.

    Sets up hooks so that at the start of every request a check
    is made that the current thread has a `TimelineCursor` inserted
    in to the db machinery, and insert one if not.

    :param timeline_factory: a callable that takes no arguments
        and returns either a `Timeline` instance, or None.
    :param prefix: the prefix to use for the entries recorded
        in the `Timeline`, or None to use the default.
    """
    inserter = TimelineCursorInserter(timeline_factory, prefix=prefix)
    inserter.connect_to_request_signals()
